# ----------------------------------------------------------------------
# |  Cache Helpers
# ----------------------------------------------------------------------
import shutil

from montyalex.fs_tools import pathexists
from montyalex.console_tools import error_stm


def remove_cache__dir(
    formatted_directory: str | bytes, silent: bool = False
) -> bool:
    if pathexists(formatted_directory):
        try:
            shutil.rmtree(formatted_directory)
            return True
        except (OSError, NotADirectoryError, IsADirectoryError):
            if not silent:
                print(
                    f"{error_stm}, Removing {formatted_directory!r}"
                    " failed in the directory given"
                )
            return False
    else:
        if not silent:
            print(
                f"Warning, Attempted to remove {formatted_directory!r},"
                " but it didn't exist"
            )
        return False



# ----------------------------------------------------------------------
# |  Cache App
# ----------------------------------------------------------------------
from functools import cache

from montyalex.typer_tools import Option, Typer
from .mtax_cache import MtaxCache
from .timecache_app import time_


# ----------------------------------------------------------------------
# |  Typer App
# ----------------------------------------------------------------------
cache_: Typer = Typer(name="cache", add_help_option=False)
cache_.add_typer(time_)


# ----------------------------------------------------------------------
# |  Typer Commands; add_, incr_, set_, get_, decr_,
# |  clear_, reset_, remove_, list_, info_
# ----------------------------------------------------------------------
@cache_.command(name="add", add_help_option=False)
def add_(
    *,
    key: str,
    value: str,
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
    timezone: str = "Etc/Greenwich",
):
    """Add a key-value pair in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.add_item((key, value), silent)


@cache
@cache_.command(name="incr", add_help_option=False)
def incr_(
    *,
    key: str,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Increment an item count in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    return mtax.incr_item(key, silent)


@cache_.command(name="set", add_help_option=False)
def set_(
    *,
    key: str,
    value: str,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Sets a key-value pair in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.set_item((key, value), silent)


@cache
@cache_.command(name="get", add_help_option=False)
def get_(
    *,
    key: str,
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Gets a key-value pair from the cache if the key exists"""
    mtax: MtaxCache = MtaxCache()
    return mtax.get_item(key, silent)


@cache
@cache_.command(name="decr", add_help_option=False)
def decr_(
    *,
    key: str,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Decrement an item count in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    return mtax.decr_item(key, silent)


@cache_.command(name="clear", add_help_option=False)
def clear_(
    *,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Clear all the key-value pairs in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.clear(silent)


@cache_.command(name="reset", add_help_option=False)
def reset_(
    *,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Reset stuck key-value pairs in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.reset(silent)


@cache_.command(name="remove", add_help_option=False)
def remove_(
    *,
    key: str,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Remove the matching key from the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.remove(key, silent)


@cache_.command(name="list", add_help_option=False)
def list_(
    *,
    timezone: str = "Etc/Greenwich",
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """List the key-value pairs stored in the cache"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.list_k_v_pairs(silent)


@cache_.command(name="info", add_help_option=False)
def info_(
    *,
    timezone: str = "Etc/Greenwich",
    list_items: bool = Option(
        False,
        "-list",
        help="List the key-value pairs stored in the cache",
    ),
    silent: bool = Option(
        False,
        "-s",
        show_default="-!s",
        help="Silence extraneous messages in console",
    ),
):
    """Shows various info and current size of caches"""
    mtax: MtaxCache = MtaxCache(timezone)
    mtax.info(list_items, silent)



# ----------------------------------------------------------------------
# |  User-Settings Tools
# ----------------------------------------------------------------------
from .usersettings import SETTINGS
from ._actions import (
    datetime_directory_action,
    simple_directory_action,
    silent_option_action,
)
from ._defaults import (
    timezone_info_default,
    fourweek_option_default,
    datetime_range_option_default,
    datetime_format_option_default,
    simple_range_option_default,
)
from ._userinfo import (
    locale_continent_userinfo,
    locale_country_userinfo,
    locale_region_userinfo,
    locale_state_userinfo,
    locale_province_userinfo,
    locale_city_userinfo,
    locale_streetname_userinfo,
    locale_streetnum_userinfo,
)


__version__ = "v1.0.0"


__all__ = [
    "SETTINGS",
    "datetime_directory_action",
    "simple_directory_action",
    "silent_option_action",
    "timezone_info_default",
    "fourweek_option_default",
    "datetime_range_option_default",
    "datetime_format_option_default",
    "simple_range_option_default",
    "locale_continent_userinfo",
    "locale_country_userinfo",
    "locale_region_userinfo",
    "locale_state_userinfo",
    "locale_province_userinfo",
    "locale_city_userinfo",
    "locale_streetname_userinfo",
    "locale_streetnum_userinfo",
    "__version__",
]



# ----------------------------------------------------------------------
# |  November
# |  TO DO by 2024/11/30
# ----------------------------------------------------------------------
from montyalex.fs_tools import (
    # abspath,  # maybe be used
    cancel,
    expanduser,
    joinpaths,
    # mkfile,  # maybe be used
    rmfile,
    # mkdirs,  # maybe be used
    # rmdirs,  # maybe be used
    # rmtree,  # maybe be used
    pathexists,
    current_working_dir as curr_workdir,  # possible renaming of variable namespace
)
from montyalex.time_tools import MtaxTime
from montyalex.typing_tools import Any
from montyalex.us_tools import SETTINGS


class MontyAlexException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class MontyAppException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class MontySettingsException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class TyperException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class TyperAppException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class CacheException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class ConsoleException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class DirectoryException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class FileSystemException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class TimeException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class UserOptionsException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


class UserSettingsException(BaseException):
    def __init__(self, message: str, value: Any) -> None:
        super().__init__(message)
        self.message = message
        self.value = value

    def __str__(self) -> str:
        return f"{self.message} {self.value}"

    def __reduce__(self) -> tuple[Any, tuple[str, Any]]:
        return self.__class__, (self.message, self.value)


USER_TIMEZONE = SETTINGS.get_value("default.info.timezone")


class TrackFile:
    def __init__(self, file_path: bool) -> None:
        self.file_path: str = file_path
        self.file_exists: bool = pathexists(file_path)
        self.created_at: str = get_mtax_time()
        self.updated_at: str = None

    def __bool__(self) -> bool:
        return self.file_exists

    def __str__(self) -> str:
        return self.file_path

    def __repr__(self) -> str:
        if self.updated_at:
            return f"File(updated_at={self.updated_at!r})"
        return f"File(created_at={self.created_at!r})"

    def updated(self) -> bool:
        return bool(self.updated_at)

    def update(self):
        self.updated_at = get_mtax_time()


def get_mtax_time():
    if USER_TIMEZONE and USER_TIMEZONE != "Notfound":
        timemod = MtaxTime(USER_TIMEZONE)
    else:
        print("No timezone found in settings, using default GMT")
        timemod = MtaxTime()
    return timemod.timestamp()


def mkfile(dirname: str, filename: str, global_: bool = False):
    directory: str = None
    if global_:
        directory = joinpaths(expanduser("~"), dirname)
    else:
        directory = joinpaths(curr_workdir, dirname)
    if not pathexists(directory):
        print("Directory does not exist")
        cancel()
    file_path = joinpaths(directory, filename)
    track: TrackFile = TrackFile(file_path)

    print(track)
    print(repr(track))
    print(bool(track))


def rmfile_(path: Any):
    rmfile(
        path
    )  #  No effect for now, potentially a rework at a later stage


def mktree():
    """Should consist of mkdirs and mkfile loops
    to create a single file, list of files, or set of files
    (A set meaning unique "lists" for each unique directory made)"""



# ----------------------------------------------------------------------
# |  Presets
# ----------------------------------------------------------------------
from enum import Enum
from rich.table import Table

from montyalex.console_tools import richconsole
from montyalex.fs_tools import cancel
from montyalex.time_tools import func_time
from montyalex.typing_tools import Any, Generator


print = richconsole.print


# ----------------------------------------------------------------------
# |  Number Base Name
# ----------------------------------------------------------------------
class BaseName(Enum):
    BINARY = 2
    TERNARY = 3
    QUATERNARY = 4
    QUINARY = 5
    SENARY = 6
    SEPTENARY = 7
    OCTAL = 8
    NONARY = 9
    DECIMAL = 10
    UNDECIMAL = 11
    DUODECIMAL = 12
    TRIDECIMAL = 13
    TETRADECIMAL = 14
    PENTADECIMAL = 15
    HEXADECIMAL = 16
    VIGESIMAL = 20
    DUOVIGESIMAL = 22
    TETRAVIGESIMAL = 24
    TRIGESIMAL = 30
    DUOTRIGESIMAL = 32
    TETRAGESIMAL = 40
    QUADRAGESIMAL = 48
    SEXAGESIMAL = 60
    TETRASEXAGESIMAL = 64
    CENTESIMAL = 100

    @classmethod
    def get_name(cls, base: int) -> str:
        for name, value in cls.__members__.items():
            if value.value == base:
                return name.lower()
        return "unknown"


# ----------------------------------------------------------------------
# |  Digit File System Tree Preset
# ----------------------------------------------------------------------
class DigitPreset:
    def __init__(self, digit: int, preset: str) -> None:
        self.digit = digit
        self.preset = preset

    def __str__(self) -> str:
        return self.preset

    def generate_base_dict(self, base: int) -> dict[str, Any]:
        chars = (
            "0123456789"
            "abcdefghijklmnopqrstuvwxyz"
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "!@#$%^&*()-=_+[]{}|;:'\",.<>?\\/"
            "±§µ¢£€¥¤©®™✓✗♥♦♣♠♪♫◊"
        )
        max_value = base**self.digit
        base_dict = {}
        for i in range(max_value):
            val = ""
            num = i
            while num > 0:
                val = chars[num % base] + val
                num //= base
            base_dict[val.zfill(self.digit)] = i
        return base_dict

    def __call__(
        self, base: int, zero_exclusive=True
    ) -> dict[str, Any]:
        base_dict = self.generate_base_dict(base)
        if zero_exclusive:
            base_dict = {
                k: v
                for k, v in base_dict.items()
                if k != ("0" * self.digit)
            }
        return base_dict

    @staticmethod
    def display_range_with_decimal(base_dict: dict) -> None:
        for base_value, decimal_value in base_dict.items():
            print(f"{base_value} = {decimal_value}")

    @staticmethod
    def range_length(
        dict_length: int, range_name: str, show_wz_range: bool = False
    ) -> tuple[str, str, str]:
        if show_wz_range:
            return (
                "(%%)",
                f"{range_name} size",
                f"[blue1]±{dict_length}[/]",
            )
        return "(++)", f"{range_name} size", f"[cyan]±{dict_length}[/]"

    def _get_base_dict(
        self, base: int, show_wz_range: bool = False
    ) -> dict:
        base_dict = self(base, not show_wz_range)
        return base_dict

    def base_dict_length(
        self, base: int, show_wz_range: bool = False
    ) -> int:
        length = base**self.digit
        if show_wz_range:
            length += 1
        return length

    def set_of_base(
        self, base: int, show_wz_range: bool = False
    ) -> set[str]:
        base_dict = self._get_base_dict(base, show_wz_range)
        return set(base_dict.keys())

    def ul_of_base(
        self, base: int, show_wz_range: bool = False
    ) -> list[str]:
        return list(self.set_of_base(base, show_wz_range))

    def ol_of_base(
        self, base: int, show_wz_range: bool = False
    ) -> list[str]:
        return sorted(self.ul_of_base(base, show_wz_range))


# ----------------------------------------------------------------------
# |  Double Digit File System Tree Preset
# ----------------------------------------------------------------------
class DoubleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(2, "Double")


# ----------------------------------------------------------------------
# |  Triple Digit File System Tree Preset (Default for Digit Presets)
# ----------------------------------------------------------------------
class TripleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(3, "Triple")


# ----------------------------------------------------------------------
# |  Quadruple Digit File System Tree Preset
# ----------------------------------------------------------------------
class QuadrupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(4, "Quadruple")


# ----------------------------------------------------------------------
# |  Quintuple Digit File System Tree Preset
# ----------------------------------------------------------------------
class QuintupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(5, "Quintuple")


# ----------------------------------------------------------------------
# |  Sextuple Digit File System Tree Preset
# ----------------------------------------------------------------------
class SextupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(6, "Sextuple")


# ----------------------------------------------------------------------
# |  Septuple Digit File System Tree Preset
# ----------------------------------------------------------------------
class SeptupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(7, "Septuple")


# ----------------------------------------------------------------------
# |  Octuple Digit File System Tree Preset
# ----------------------------------------------------------------------
class OctupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(8, "Octuple")


# ----------------------------------------------------------------------
# |  Nonuple Digit File System Tree Preset
# ----------------------------------------------------------------------
class NonupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(9, "Nonuple")


# ----------------------------------------------------------------------
# |  Decuple Digit File System Tree Preset
# ----------------------------------------------------------------------
class DecupleDigitPreset(DigitPreset):
    def __init__(self) -> None:
        super().__init__(10, "Decuple")


# ----------------------------------------------------------------------
# |  File System Tree Presets
# ----------------------------------------------------------------------
class Preset:
    digit_default_dict = {
        "single": DigitPreset(1, "Single"),
        "double": DoubleDigitPreset(),
        "triple": TripleDigitPreset(),
        "quadruple": QuadrupleDigitPreset(),
        "quintuple": QuintupleDigitPreset(),
        "sextuple": SextupleDigitPreset(),
        "septuple": SeptupleDigitPreset(),
        "octuple": OctupleDigitPreset(),
        "nonuple": NonupleDigitPreset(),
        "decuple": DecupleDigitPreset(),
    }

    def __init__(
        self,
        kind: str,
        *,
        name: str = None,
        formatted: str | bool = False,
        numbase: int = 10,
    ) -> None:
        self.kind = kind

        self.name = name
        self.format = formatted
        self.numbase = numbase
        self.preset: DigitPreset | object = None

        if self.kind == "digit":
            if self.name is None:
                self.name = "triple"
            self.preset = self.digit_default_dict[self.name]

    def __len__(self) -> int:
        if self.kind == "digit":
            return self.preset.base_dict_length(self.numbase, False)
        return 0

    def _get_base_dict(
        self, base: int, show_wz_range: bool = False
    ) -> dict:
        base_dict = self.digit_default_dict[self.name](
            base, not show_wz_range
        )
        return base_dict

    def _table_items_generator(
        self, show_wz_range: bool
    ) -> Generator[tuple[dict, int], Any, None]:
        base_dict = self._get_base_dict(self.numbase, show_wz_range)
        yield base_dict

    @func_time("took; to generate")
    def table(self, show_wz_range: bool = False) -> Table:
        target_base_name = BaseName.get_name(self.numbase)
        rows = []
        for base_dict in self._table_items_generator(show_wz_range):
            rows.append(
                self.preset.range_length(
                    len(base_dict), target_base_name, show_wz_range
                )
            )
        t = Table(
            show_header=False,
            pad_edge=False,
            highlight=True,
            title=f"{self.name.capitalize()} Digit Range",
            title_style="bold turquoise2",
            title_justify="left",
        )
        t.add_column("1", no_wrap=True)
        t.add_column("2", no_wrap=True)
        t.add_column("3", justify="right", no_wrap=True)
        for row in rows:
            t.add_row(*row)
        return t


# ----------------------------------------------------------------------
# |  Global Specified File System Tree Preset Objects
# ----------------------------------------------------------------------
binary_single = Preset("digit", name="single", numbase=2)
binary_double = Preset("digit", name="double", numbase=2)
binary_triple = Preset("digit", name="triple", numbase=2)
binary_quadruple = Preset("digit", name="quadruple", numbase=2)
binary_quintuple = Preset("digit", name="quintuple", numbase=2)
binary_sextuple = Preset("digit", name="sextuple", numbase=2)
binary_septuple = Preset("digit", name="septuple", numbase=2)
binary_octuple = Preset("digit", name="octuple", numbase=2)
binary_nonuple = Preset("digit", name="nonuple", numbase=2)
binary_decuple = Preset("digit", name="decuple", numbase=2)

ternary_single = Preset("digit", name="single", numbase=3)
ternary_double = Preset("digit", name="double", numbase=3)
ternary_triple = Preset("digit", name="triple", numbase=3)
ternary_quadruple = Preset("digit", name="quadruple", numbase=3)
ternary_quintuple = Preset("digit", name="quintuple", numbase=3)
ternary_sextuple = Preset("digit", name="sextuple", numbase=3)
ternary_septuple = Preset("digit", name="septuple", numbase=3)
ternary_octuple = Preset("digit", name="octuple", numbase=3)
ternary_nonuple = Preset("digit", name="nonuple", numbase=3)
ternary_decuple = Preset("digit", name="decuple", numbase=3)

quaternary_single = Preset("digit", name="single", numbase=4)
quaternary_double = Preset("digit", name="double", numbase=4)
quaternary_triple = Preset("digit", name="triple", numbase=4)
quaternary_quadruple = Preset("digit", name="quadruple", numbase=4)
quaternary_quintuple = Preset("digit", name="quintuple", numbase=4)
quaternary_sextuple = Preset("digit", name="sextuple", numbase=4)
quaternary_septuple = Preset("digit", name="septuple", numbase=4)
quaternary_octuple = Preset("digit", name="octuple", numbase=4)
quaternary_nonuple = Preset("digit", name="nonuple", numbase=4)
quaternary_decuple = Preset("digit", name="decuple", numbase=4)

quinary_single = Preset("digit", name="single", numbase=5)
quinary_double = Preset("digit", name="double", numbase=5)
quinary_triple = Preset("digit", name="triple", numbase=5)
quinary_quadruple = Preset("digit", name="quadruple", numbase=5)
quinary_quintuple = Preset("digit", name="quintuple", numbase=5)
quinary_sextuple = Preset("digit", name="sextuple", numbase=5)
quinary_septuple = Preset("digit", name="septuple", numbase=5)
quinary_octuple = Preset("digit", name="octuple", numbase=5)
quinary_nonuple = Preset("digit", name="nonuple", numbase=5)

senary_single = Preset("digit", name="single", numbase=6)
senary_double = Preset("digit", name="double", numbase=6)
senary_triple = Preset("digit", name="triple", numbase=6)
senary_quadruple = Preset("digit", name="quadruple", numbase=6)
senary_quintuple = Preset("digit", name="quintuple", numbase=6)
senary_sextuple = Preset("digit", name="sextuple", numbase=6)
senary_septuple = Preset("digit", name="septuple", numbase=6)
senary_octuple = Preset("digit", name="octuple", numbase=6)
senary_nonuple = Preset("digit", name="nonuple", numbase=6)

septenary_single = Preset("digit", name="single", numbase=7)
septenary_double = Preset("digit", name="double", numbase=7)
septenary_triple = Preset("digit", name="triple", numbase=7)
septenary_quadruple = Preset("digit", name="quadruple", numbase=7)
septenary_quintuple = Preset("digit", name="quintuple", numbase=7)
septenary_sextuple = Preset("digit", name="sextuple", numbase=7)
septenary_septuple = Preset("digit", name="septuple", numbase=7)
septenary_octuple = Preset("digit", name="octuple", numbase=7)

octal_single = Preset("digit", name="single", numbase=8)
octal_double = Preset("digit", name="double", numbase=8)
octal_triple = Preset("digit", name="triple", numbase=8)
octal_quadruple = Preset("digit", name="quadruple", numbase=8)
octal_quintuple = Preset("digit", name="quintuple", numbase=8)
octal_sextuple = Preset("digit", name="sextuple", numbase=8)
octal_septuple = Preset("digit", name="septuple", numbase=8)

nonary_single = Preset("digit", name="single", numbase=9)
nonary_double = Preset("digit", name="double", numbase=9)
nonary_triple = Preset("digit", name="triple", numbase=9)
nonary_quadruple = Preset("digit", name="quadruple", numbase=9)
nonary_quintuple = Preset("digit", name="quintuple", numbase=9)
nonary_sextuple = Preset("digit", name="sextuple", numbase=9)
nonary_septuple = Preset("digit", name="septuple", numbase=9)

decimal_single = Preset("digit", name="single")
decimal_double = Preset("digit", name="double")
decimal_triple = Preset("digit", name="triple")
decimal_quadruple = Preset("digit", name="quadruple")
decimal_quintuple = Preset("digit", name="quintuple")
decimal_sextuple = Preset("digit", name="sextuple")
decimal_septuple = Preset("digit", name="septuple")
decimal_octuple = Preset("digit", name="octuple")
decimal_nonuple = Preset("digit", name="nonuple")
decimal_decuple = Preset("digit", name="decuple")

duodecimal_single = Preset("digit", name="single", numbase=12)
duodecimal_double = Preset("digit", name="double", numbase=12)
duodecimal_triple = Preset("digit", name="triple", numbase=12)
duodecimal_quadruple = Preset("digit", name="quadruple", numbase=12)
duodecimal_quintuple = Preset("digit", name="quintuple", numbase=12)
duodecimal_sextuple = Preset("digit", name="sextuple", numbase=12)

hexadecimal_single = Preset("digit", name="single", numbase=16)
hexadecimal_double = Preset("digit", name="double", numbase=16)
hexadecimal_triple = Preset("digit", name="triple", numbase=16)
hexadecimal_quadruple = Preset("digit", name="quadruple", numbase=16)
hexadecimal_quintuple = Preset("digit", name="quintuple", numbase=16)

base32_single = Preset("digit", name="single", numbase=32)
base32_double = Preset("digit", name="double", numbase=32)
base32_triple = Preset("digit", name="triple", numbase=32)
base32_quadruple = Preset("digit", name="quadruple", numbase=32)

base64_single = Preset("digit", name="single", numbase=64)
base64_double = Preset("digit", name="double", numbase=64)
base64_triple = Preset("digit", name="triple", numbase=64)


# ----------------------------------------------------------------------
# |  Least Possible Range
# ----------------------------------------------------------------------
LPR: dict[str, Preset] = {
    "single": binary_single,
    "double": binary_double,
    "triple": binary_triple,
    "quadruple": binary_quadruple,
    "quintuple": binary_quintuple,
    "sextuple": binary_sextuple,
    "septuple": binary_septuple,
    "octuple": binary_octuple,
    "nonuple": binary_nonuple,
    "decuple": binary_decuple,
}


# ----------------------------------------------------------------------
# |  Greatest Possible Range
# ----------------------------------------------------------------------
GPR: dict[str, Preset] = {
    "single": base64_single,
    "double": base64_double,
    "triple": base64_triple,
    "quadruple": base32_quadruple,
    "quintuple": hexadecimal_quintuple,
    "sextuple": duodecimal_sextuple,
    "septuple": nonary_septuple,
    "octuple": septenary_octuple,
    "nonuple": senary_nonuple,
    "decuple": quaternary_decuple,
}


# ----------------------------------------------------------------------
# |  Most Common Range
# ----------------------------------------------------------------------
MCR: dict[str, Preset] = {
    "single": decimal_single,
    "double": decimal_double,
    "triple": decimal_triple,
    "quadruple": decimal_quadruple,
    "quintuple": decimal_quintuple,
    "sextuple": decimal_sextuple,
    "septuple": decimal_septuple,
    "octuple": decimal_octuple,
    "nonuple": decimal_nonuple,
    "decuple": decimal_decuple,
}


# ---------------------------------------------------------------------------
# |  Example Usage
# ---------------------------------------------------------------------------
if __name__ == "__main__":

    def main():
        digits = (
            "(single|double|triple|quadruple|quintuple|"
            "sextuple|septuple|octuple|nonuple|decuple)"
        )
        digits_input = (
            input(f"How many digits should be used? {digits}: ")
            .lower()
            .strip()
        )
        base = input(
            "Which base should be used? (10:default|2-64): "
        ).strip()
        if base == "" or not base.isnumeric():
            base = 10
        else:
            base = int(base)
            if base > 64 or base < 2:
                print(f"Base {base} not allowed")
                cancel()

        double_up_input = (
            input(
                "Do you want to double the folder name? (a=aa, y|N): "
            )
            .lower()
            .strip()
        )
        zeroith_input = (
            input("Do you want to include a zeroith folder? (y|N): ")
            .lower()
            .strip()
        )
        _preset = MCR[f"{digits_input}"].preset.ol_of_base(
            base, bool(zeroith_input == "y")
        )
        for b in _preset:
            print(b)
            if double_up_input == "y":
                print(f"{b}_{b}")
        if double_up_input == "y":
            print(f"len({len(_preset) * 2})")
        else:
            print(f"len({len(_preset)})")



# ----------------------------------------------------------------------
# |  Tree-Id
# ----------------------------------------------------------------------
from montyalex.fs_tools import (
    current_working_dir,
    joinpaths,
    pathexists,
    mkdirs,
    rmfile,
)
from montyalex.object_tools import singleton, pdumps, ploads


# ----------------------------------------------------------------------
# |  Singleton File System Tree
# ----------------------------------------------------------------------
@singleton(get_identifier=lambda name: name)
class FileSystemTree:
    def __init__(self, name: str):
        self.name: str = name
        self.age = 0
        self.template: str = None
        self.nodes = [0, 0, 0]
        self._path: str = joinpaths(
            current_working_dir, ".mtax", ".pkl", f"tree-{self.name}"
        )
        self.exists: bool = pathexists(self._path)

    def __str__(self) -> str:
        return f"{self.__dict__}"

    def progress(self):
        self.age += 1
        self.write()

    def write(self) -> None:
        mkdirs(
            self._path.removesuffix(f"tree-{self.name}"), exist_ok=True
        )
        with open(self._path, "wb") as tree_pkl:
            tree_pkl.write(pdumps(self.__dict__))

    def setup(self, *, template: str = None):
        if template:
            self.template = template

    def scan(self):
        if self.exists:
            with open(self._path, "rb") as tree_pkl:
                return ploads(tree_pkl.read())
        return self.__dict__

    def delete(self):
        if self.exists:
            rmfile(self._path)


# ---------------------------------------------------------------------------
# |  Example Usage
# ---------------------------------------------------------------------------
if __name__ == "__main__":
    tree1 = FileSystemTree("name")
    print(tree1)
    print(tree1.scan())
